/* Generated by Font Squirrel (http://www.fontsquirrel.com) on April 19, 2013 06:38:14 PM America/New_York */

@font-face {
    font-family: 'SourceCodeProRegular';
    src: url('SourceCodePro-Regular-webfont.eot');
    src: url('SourceCodePro-Regular-webfont.eot?#iefix') format('embedded-opentype'),
         url('SourceCodePro-Regular-webfont.woff') format('woff'),
         url('SourceCodePro-Regular-webfont.ttf') format('truetype'),
         url('SourceCodePro-Regular-webfont.svg#SourceCodeProRegular') format('svg');
    font-weight: normal;
    font-style: normal;

}
