// std stuff
#include <map>
#include <string>
#include <vector>
#include <set>
#include <list>
#include <iostream>
#include <limits>
#include <sstream>
#include <fstream>
#include <iomanip>
#include <tuple>
#include <algorithm>
#include <cassert>
#include <climits>
#include <iosfwd>
#include <memory>
#include <type_traits>
#include <map>
#include <unordered_map>
#include <chrono>

// boost stuff
#include <boost/chrono.hpp>
#include <boost/variant.hpp>
#include <boost/math/distributions/normal.hpp>
#include <boost/algorithm/string.hpp>
#include <boost/lexical_cast.hpp>

#include <nonius/nonius.h++>
#include <nonius/html_group_reporter.h>
